//*****************************************************************************
//*****************************************************************************
//  FILENAME:   sensor.c
//  Version: 1.0, 2007/03/08
//  Generated by Andy Kiser
//
//  DESCRIPTION:  sensor specific functions 
//
//-----------------------------------------------------------------------------
//  Copyright (c) Toradex AG 2007. All Rights Reserved.
//*****************************************************************************
//*****************************************************************************
//#define __DEBUG__
#include <m8c.h>        // part specific constants and macros
#include "sensor.h"
#include "psocgpioint.h"
#include "psocapi.h"
#include "feature.h"
#include "parameters.h"
#include "types_n_macros.h"
//----------------------------------------
// Sensor parameters
// Factory Defaults are defined here (NUM_xxx are defined in "sensor.h")
//----------------------------------------
const BYTE FactoryDefault_Param_8   [NUM_PARAM_8] =
{
	0x00,			// report mode
	0x00,			// LED mode
	GP2D12J0000F	// Sensor Type is GP2D12J0000F (80cm)
};

const WORD FactoryDefault_Param_16  [NUM_PARAM_16] = 
{
	500,		// report rate [ms]
	500,		// sample rate [ms]
	50000		// sample rate [us]	(not used for this sensor)
};

const DWORD FactoryDefault_Param_32  [NUM_PARAM_32] =
{
	0x00000000	// currently no 32bit parameters
};

const tParam_168 FactoryDefault_Param_168 [NUM_PARAM_168] =
{
	"Distance Sensor",	// user device name
	"Frame Number",		// user name Channel 0
	"Distance"			// user name Channel 1
};


//----------------------------------------
// this module's global variables
//----------------------------------------
tReport Report;

//----------------------------------------
// constant lookup-table for ADC->physical value conversion
//----------------------------------------

//----------------------------------------
// type definitions
//----------------------------------------

typedef enum 
{	
	eReset,
	eIdle,
	eReadADC,
	eCalculating
} tFSM_Measure;

//----------------------------------------
// variables

tFSM_Measure	FSM;
WORD			Countdown_1ms;

BOOL			NewSample = FALSE;

// store original drivemodes during suspend
BYTE			DM0_2, DM0_1, DM0_0;
BYTE			DM2_2, DM2_1, DM2_0;
BYTE			ARF;

//----------------------------------------
// function prototypes
static int CalculateDistance (WORD ADC);

//-----------------------------------------------------------------------------
// Initializes the Sensor 
//-----------------------------------------------------------------------------
void Sensor_Start ()
{
	// turn on power to the sensor
	GPIO_LOW(OFF_nON)
	
	// start ADC sampling
	PGA_Start (PGA_HIGHPOWER);
	DelSig_Start (DelSig_MEDPOWER);
	DelSig_StartAD(); 		// Start continous sampling
	
	FSM = eReset;
}
//-----------------------------------------------------------------------------
// Prepare Sensor for Suspend
//-----------------------------------------------------------------------------
void Sensor_Suspend()
{
	// stop and power off ADC
	DelSig_StopAD();
	DelSig_Stop();
	PGA_Stop();
	// turn off power for the sensor
	GPIO_HIGH(OFF_nON)

	// store original drivemodes during  suspend
	DM0_2 = PRT0DM2;
	DM0_1 = PRT0DM1;
	DM0_0 = PRT0DM0;
	DM2_2 = PRT2DM2;
	DM2_1 = PRT2DM1;
	DM2_0 = PRT2DM0;
	// store Analog power settings
	ARF   = ARF_CR;
	
	// set all I/Os to/from the sensor to High-Z Analog (=110), 
	// except the OFF_nON signal. 
	// OFF_nON is on Port0, all other signals on Port2
	PRT0DM2  = 0xFF;
	PRT0DM1  = 0xFF;
	PRT0DM0  = 0x00;
	PRT2DM2 |= ~OFF_nON_MASK;
	PRT2DM1 |= ~OFF_nON_MASK;
	PRT2DM0 &=  OFF_nON_MASK;

}
//-----------------------------------------------------------------------------
// Resume Sensor from Suspend mode
//-----------------------------------------------------------------------------
void Sensor_Resume()
{
	BYTE ReadMode;
	
	// restore drivemodes for GPIOs to/from the sensor
	PRT2DM2 = DM2_2;
	PRT2DM1 = DM2_1;
	PRT2DM0 = DM2_0;

	PRT0DM2 = DM0_2;
	PRT0DM1 = DM0_1;
	PRT0DM0 = DM0_0;
	
	ARF_CR = ARF;

	// turn on power for the sensor
	Sensor_Start();
	FSM = eReset;
}


//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// cannot be handled with the standard functionality in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//
//	return Value: number of Parameter bytes (1-31) that are 
//				  returned in the IN report
//-----------------------------------------------------------------------------
BYTE Sensor_Specific_Feature(void)
{
	// currently no special feature reports are handled
	return 0;
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// sets / gets a 8/16/32/168bit parameter in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//-----------------------------------------------------------------------------
void Sensor_Specific_Parameter(BYTE *NumberOfReturnedBytes)
{
	// currently no specific parameters are handled
}

//-----------------------------------------------------------------------------
// Read ADC 
//-----------------------------------------------------------------------------
static WORD GP2D12_ReadADC (void)
{
	WORD Data;

	if ( DelSig_fIsDataAvailable() )
	{
		M8C_DisableGInt;
		Data = DelSig_wGetDataClearFlag();	
		M8C_EnableGInt;
		
		return Data;
	}
	else
		return 0xFFFF;
}

//-----------------------------------------------------------------------------
// void Sensor_FSM_Measure(void)
//
//	finite state machine to control the measurement process for 
//  the Sensor
//
//  IMPORTANT: 	call this routine once every millisecond. Otherwise the timing
//				parameters cannot be met.
//-----------------------------------------------------------------------------
void Sensor_FSM_Measure() 
{
	static WORD ADC; // ADC value
	WORD 		Distance;
	BYTE		SensorType;
	
	// decrement millisecond counter
	if (Countdown_1ms != 0)
		Countdown_1ms--;

	// process actual state machine
	switch (FSM)
	{
		case eReset:
			Countdown_1ms = 10; // 10ms delay until first measurement finished
			FSM = eIdle;
			break;
			
		case eIdle: 
			if (Countdown_1ms == 0)
			{
				Get_Param_16(PAR16_SAMPLERATE_MS, VOLATILE, &Countdown_1ms);;
				FSM = eReadADC;					  
			}
			break;			
			
		case eReadADC:
			Get_Param_8(PAR8_SENSORTYPE, VOLATILE, &SensorType);

			switch(SensorType)
			{
				case GP2Y0A41SK0F: // 30cm type
					PGA_SetGain(PGA_G0_81); // gain = 0.813
					break;
					
				case GP2D12J0000F: // 80cm type
					PGA_SetGain(PGA_G1_00);	// gain = 1.000
					break;
				
				case 0xFF:	// debug with 80cm type
					PGA_SetGain(PGA_G1_00);	// gain = 1.000
					break;
			}		
			ADC = GP2D12_ReadADC();
			FSM = eCalculating;
			break;			

		case eCalculating:
			Distance = CalculateDistance (ADC);

			Report.Ch1_LSB = Distance & 0xFF;
			Report.Ch1_MSB = Distance / 0x100;

//			Report.Ch1_LSB = ADC & 0xFF;
//			Report.Ch1_MSB = ADC / 0x100;

			NewSample = TRUE;
			
			FSM = eIdle;
			break;			
	}
}

//-----------------------------------------------------------------------------
// Routine: int CalculateDistance (WORD ADC)
//
// Description: Calculate the distance from the ADC output of a IR distance sensor
//
// Return: WORD - distance [0.1mm]
//
//-----------------------------------------------------------------------------
extern const WORD ADC2L_80cm[513-64];	// for GP2D12J0000F
extern const WORD ADC2L_30cm[513];		// for GP2Y0A41SK0F 

static int CalculateDistance (WORD ADC)
{
	BYTE SensorType;
	WORD index;
	WORD L_Below, L_Above, sum = 0;
	
	Get_Param_8(PAR8_SENSORTYPE, VOLATILE, &SensorType);
	
	switch(SensorType)
	{
		case GP2Y0A41SK0F: // 30cm type
			// calculate index into lookup table
			index = (ADC -0) >> 3;	
			L_Below = ADC2L_30cm[index];
			L_Above = ADC2L_30cm[index+1];
			break;
			
		case GP2D12J0000F: // 80cm type
			if (ADC <=512)
				index = 0;
			else
				index = (ADC-512) >>  3;
				
			// do linear interpolation
			L_Below = ADC2L_80cm[index];
			L_Above = ADC2L_80cm[index+1];
			break;
			
		case 0xFF:	// Debug, Raw ADC value
			L_Below = ADC;
			L_Above = ADC+8;
			break;
	}
	
	// do linear interpolation, assuming every 8th data is in the lookup table
	sum = L_Below;								// sum ~ L
	sum += (ADC & 0x0001) ? L_Above : L_Below;  // sum ~ 2xL

	L_Above <<=1; L_Below <<=1;
	sum += (ADC & 0x0002) ? L_Above : L_Below;	// sum ~ 4xL
	sum >>=1;									// sum ~ 2xL
	sum += (ADC & 0x0004) ? L_Above : L_Below;	// sum ~ 4xL
	
	return sum >>2;
}


const WORD ADC2L_80cm[] =	// table lists every 8th distance value, starting at ADC=512
{
	10899,// ADC=512  =>  L=1089.9mm
	10702,// ADC=520  => L=1070.2mm
	10511,// 
	10327,
	10148,
	9975,
	9807,
	9645,
	9487,
	9334,
	9186,
	9042,
	8902,
	8766,
	8634,
	8505,
	8380,
	8258,
	8140,
	8025,
	7912,
	7803,
	7696,
	7592,
	7490,
	7391,
	7295,
	7200,
	7108,
	7018,
	6931,
	6845,
	6761,
	6679,
	6599,
	6520,
	6444,
	6368,
	6295,
	6223,
	6153,
	6084,
	6016,
	5950,
	5885,
	5822,
	5759,
	5698,
	5638,
	5580,
	5522,
	5466,
	5410,
	5356,
	5302,
	5250,
	5198,
	5148,
	5098,
	5049,
	5002,
	4954,
	4908,
	4863,
	4818,
	4774,
	4731,
	4688,
	4646,
	4605,
	4565,
	4525,
	4486,
	4447,
	4409,
	4372,
	4335,
	4299,
	4264,
	4228,
	4194,
	4160,
	4126,
	4093,
	4061,
	4029,
	3997,
	3966,
	3935,
	3905,
	3875,
	3846,
	3817,
	3788,
	3760,
	3732,
	3705,
	3677,
	3651,
	3624,
	3598,
	3573,
	3547,
	3522,
	3498,
	3473,
	3449,
	3425,
	3402,
	3379,
	3356,
	3333,
	3311,
	3289,
	3267,
	3246,
	3225,
	3204,
	3183,
	3163,
	3142,
	3122,
	3103,
	3083,
	3064,
	3045,
	3026,
	3007,
	2989,
	2970,
	2952,
	2935,
	2917,
	2900,
	2882,
	2865,
	2848,
	2832,
	2815,
	2799,
	2783,
	2767,
	2751,
	2735,
	2720,
	2704,
	2689,
	2674,
	2659,
	2645,
	2630,
	2616,
	2602,
	2587,
	2573,
	2560,
	2546,
	2532,
	2519,
	2506,
	2493,
	2479,
	2467,
	2454,
	2441,
	2429,
	2416,
	2404,
	2392,
	2380,
	2368,
	2356,
	2344,
	2332,
	2321,
	2309,
	2298,
	2287,
	2276,
	2265,
	2254,
	2243,
	2232,
	2221,
	2211,
	2200,
	2190,
	2180,
	2170,
	2160,
	2150,
	2140,
	2130,
	2120,
	2110,
	2101,
	2091,
	2082,
	2072,
	2063,
	2054,
	2045,
	2036,
	2027,
	2018,
	2009,
	2000,
	1992,
	1983,
	1974,
	1966,
	1958,
	1949,
	1941,
	1933,
	1925,
	1916,
	1908,
	1900,
	1893,
	1885,
	1877,
	1869,
	1862,
	1854,
	1846,
	1839,
	1831,
	1824,
	1817,
	1809,
	1802,
	1795,
	1788,
	1781,
	1774,
	1767,
	1760,
	1753,
	1746,
	1740,
	1733,
	1726,
	1720,
	1713,
	1706,
	1700,
	1694,
	1687,
	1681,
	1675,
	1668,
	1662,
	1656,
	1650,
	1644,
	1638,
	1632,
	1626,
	1620,
	1614,
	1608,
	1602,
	1596,
	1591,
	1585,
	1579,
	1574,
	1568,
	1563,
	1557,
	1552,
	1546,
	1541,
	1535,
	1530,
	1525,
	1519,
	1514,
	1509,
	1504,
	1499,
	1494,
	1489,
	1484,
	1479,
	1474,
	1469,
	1464,
	1459,
	1454,
	1449,
	1444,
	1440,
	1435,
	1430,
	1425,
	1421,
	1416,
	1412,
	1407,
	1403,
	1398,
	1394,
	1389,
	1385,
	1380,
	1376,
	1371,
	1367,
	1363,
	1359,
	1354,
	1350,
	1346,
	1342,
	1338,
	1333,
	1329,
	1325,
	1321,
	1317,
	1313,
	1309,
	1305,
	1301,
	1297,
	1293,
	1289,
	1286,
	1282,
	1278,
	1274,
	1270,
	1267,
	1263,
	1259,
	1255,
	1252,
	1248,
	1244,
	1241,
	1237,
	1234,
	1230,
	1227,
	1223,
	1220,
	1216,
	1213,
	1209,
	1206,
	1202,
	1199,
	1196,
	1192,
	1189,
	1186,
	1182,
	1179,
	1176,
	1172,
	1169,
	1166,
	1163,
	1160,
	1156,
	1153,
	1150,
	1147,
	1144,
	1141,
	1138,
	1135,
	1132,
	1128,
	1125,
	1122,
	1119,
	1116,
	1114,
	1111,
	1108,
	1105,
	1102,
	1099,
	1096,
	1093,
	1090,
	1088,
	1085,
	1082,
	1079,
	1076,
	1074,
	1071,
	1068,
	1065,
	1063,
	1060,
	1057,
	1055,
	1052,
	1049,
	1047,
	1044,
	1041,
	1039,
	1036,
	1034,
	1031,
	1028,
	1026,
	1023,
	1021,
	1018,
	1016,
	1013,
	1011,
	1008,
	1006,
	1004,
	1001,
	999,
	996,
	994,
	991,
	989,
	987,
	984,
	982,
	980,
	977,
	975,
	973,
	970,
	968,
	966,
	964,
	961,
	959,
	957,
	955,
	952,
	950,
	948,
	946,// 
	944,// ADC=4088  => L=94.4mm
	941// ADC=4096 (does not exist)  => L=94.1mm
};

const WORD ADC2L_30cm[513] =// table lists every 8th distance value, starting at ADC=512
{
9521,// ADC=0  =>  L=952.1mm
9362,// ADC=8  => L=936.2mm
9205,// 
9050,
8896,
8744,
8595,
8447,
8301,
8157,
8015,
7874,
7736,
7599,
7464,
7332,
7201,
7071,
6944,
6819,
6695,
6574,
6454,
6336,
6220,
6106,
5994,
5884,
5775,
5669,
5564,
5461,
5360,
5261,
5164,
5069,
4975,
4884,
4794,
4706,
4620,
4536,
4454,
4373,
4295,
4218,
4144,
4071,
4000,
3931,
3864,
3798,
3735,
3673,
3614,
3556,
3500,
3446,
3394,
3345,
3298,
3254,
3212,
3171,
3130,
3089,
3049,
3010,
2971,
2933,
2897,
2862,
2829,
2797,
2766,
2737,
2708,
2680,
2652,
2623,
2594,
2565,
2536,
2507,
2478,
2450,
2421,
2393,
2364,
2336,
2307,
2279,
2250,
2222,
2194,
2167,
2142,
2117,
2093,
2071,
2049,
2029,
2009,
1991,
1973,
1957,
1942,
1929,
1917,
1907,
1898,
1890,
1882,
1872,
1862,
1850,
1837,
1823,
1808,
1792,
1776,
1758,
1740,
1721,
1701,
1681,
1661,
1643,
1626,
1610,
1594,
1579,
1565,
1551,
1538,
1526,
1514,
1503,
1493,
1483,
1473,
1464,
1454,
1444,
1433,
1423,
1413,
1403,
1392,
1382,
1372,
1362,
1353,
1343,
1334,
1325,
1315,
1307,
1298,
1289,
1280,
1272,
1263,
1255,
1247,
1238,
1230,
1222,
1214,
1206,
1198,
1190,
1182,
1175,
1167,
1160,
1152,
1145,
1138,
1131,
1124,
1117,
1110,
1103,
1097,
1090,
1083,
1077,
1071,
1064,
1058,
1052,
1046,
1040,
1034,
1029,
1023,
1017,
1012,
1006,
1001,
995,
990,
985,
980,
975,
970,
965,
961,
956,
952,
948,
943,
939,
935,
931,
927,
923,
918,
914,
910,
906,
902,
898,
894,
890,
886,
882,
879,
875,
871,
868,
864,
861,
857,
854,
850,
847,
844,
841,
837,
834,
831,
828,
824,
821,
818,
814,
811,
808,
805,
801,
798,
795,
791,
788,
785,
782,
778,
775,
772,
769,
765,
762,
759,
756,
753,
749,
746,
743,
740,
737,
734,
731,
728,
725,
722,
719,
716,
713,
710,
707,
704,
701,
698,
695,
693,
690,
687,
684,
682,
679,
676,
674,
671,
668,
666,
663,
661,
658,
656,
653,
651,
648,
646,
643,
641,
638,
636,
634,
631,
629,
627,
624,
622,
619,
617,
615,
612,
610,
608,
605,
603,
601,
599,
596,
594,
592,
590,
587,
585,
583,
581,
579,
576,
574,
572,
570,
568,
566,
564,
562,
559,
557,
555,
553,
551,
549,
547,
545,
543,
541,
539,
537,
536,
534,
532,
530,
528,
526,
524,
522,
520,
518,
516,
515,
513,
511,
509,
507,
505,
504,
502,
500,
498,
496,
495,
493,
491,
489,
488,
486,
484,
483,
481,
479,
478,
476,
474,
473,
471,
470,
468,
466,
465,
463,
462,
460,
459,
457,
456,
454,
453,
452,
450,
449,
447,
446,
445,
443,
442,
440,
439,
438,
436,
435,
433,
432,
431,
429,
428,
427,
425,
424,
422,
421,
420,
418,
417,
416,
414,
413,
411,
410,
409,
407,
406,
404,
403,
402,
400,
399,
398,
396,
395,
394,
393,
392,
392,
391,
391,
391,
391,
391,
391,
391,
392,
393,
393,
394,
395,
395,
396,
396,
395,
395,
394,
392,
391,
389,
386,
383,
380,
377,
373,
368,
364,
359,
354,
349,
343,
337,
331,
325,
319,
312,
305,
298,
291,
283,
276,
268,
260,
251,
243,
234,
225,
216,
207,
197,
187,
177,
167,
156,
146,
135,
124,
113,
101,
89,
77,// 
65,// ADC=4088  => L=0.65mm
53// ADC=4096 (does not exist)  => L=0.53mm
};

//-----------------------------------------------------------------------------
void Sensor_Debug(void)
{
/* currently no debug function */
}